@@ -63,6 +63,7 @@ ISOTREE_EXPORTED int fit_iforest(IsoForest *model_outputs, ExtIsoForest *model_o
                 size_t nrows, size_t sample_size, size_t ntrees,
                 size_t max_depth,   size_t ncols_per_tree,
                 bool   limit_depth, bool penalize_range, bool standardize_data,
+                ScoringMetric scoring_metric,
                 bool   standardize_dist, double tmat[],
                 double output_depths[], bool standardize_depth,
                 real_t col_weights[], bool weigh_by_kurt,
@@ -85,6 +86,7 @@ ISOTREE_EXPORTED int fit_iforest(IsoForest *model_outputs, ExtIsoForest *model_o
                 nrows, sample_size, ntrees,
                 max_depth,   ncols_per_tree,
                 limit_depth, penalize_range, standardize_data,
+                scoring_metric,
                 standardize_dist, tmat,
                 output_depths, standardize_depth,
                 col_weights, weigh_by_kurt,
